# -*- coding: utf-8 -*-
"""Remove all documents in a search index that no longer exist in source queryset."""
from elasticsearch_django.management.commands import BaseSearchCommand
from elasticsearch_django.index import prune_index


class Command(BaseSearchCommand):

    """Run the management command."""

    help = "Remove all out-of-date documents in a search index."
    description = "Prune search index"

    def do_index_command(self, index, **options):
        """Prune search index."""
        return prune_index(index)



# -*- coding: utf-8 -*-
"""Search3 index-related functions."""
import logging

from elasticsearch import helpers

from .settings import (
    get_setting,
    get_index_mapping,
    get_index_models,
    get_client
)

logger = logging.getLogger(__name__)


def create_index(index):
    """Create an index and apply mapping if appropriate."""
    logger.info("Creating search index: '%s'", index)
    client = get_client()
    return client.indices.create(index=index, body=get_index_mapping(index))


def update_index(index):
    """Re-index every document in a named index."""
    logger.info("Updating search index: '%s'", index)
    client = get_client()
    responses = []
    for model in get_index_models(index):
        logger.info("Updating search index model: '%s'", model.search_doc_type)
        objects = model.objects.get_search_queryset(index).iterator()
        actions = bulk_actions(objects, index=index, action='index')
        response = helpers.bulk(client, actions, chunk_size=get_setting('chunk_size'))
        responses.append(response)
    return responses


def delete_index(index):
    """Delete index entirely (removes all documents and mapping)."""
    logger.info("Deleting search index: '%s'", index)
    client = get_client()
    return client.indices.delete(index=index)


def prune_index(index):
    """Remove all orphaned documents from an index.

    This function works by scanning the remote index, and in each returned
    batch of documents looking up whether they appear in the default index
    queryset. If they don't (they've been deleted, or no longer fit the qs
    filters) then they are deleted from the index. The deletion is done in
    one hit after the entire remote index has been scanned.

    The elasticsearch.helpers.scan function returns each document one at a
    time, so this function can swamp the database with SELECT requests.

    Please use sparingly.

    Returns a list of ids of all the objects deleted.

    """
    logger.info("Pruning missing objects from index '%s'", index)
    prunes = []
    responses = []
    client = get_client()
    for model in get_index_models(index):
        for hit in scan_index(index, model):
            obj = _prune_hit(hit, model)
            if obj:
                prunes.append(obj)
        logger.info(
            "Found %s objects of type '%s' for deletion from '%s'.",
            len(prunes), model, index
        )
        if len(prunes) > 0:
            actions = bulk_actions(prunes, index, 'delete')
            response = helpers.bulk(client, actions, chunk_size=get_setting('chunk_size'))
            responses.append(response)
    return responses


def _prune_hit(hit, model):
    """
    Check whether a document should be pruned.

    This method uses the SearchDocumentManagerMixin.in_search_queryset method
    to determine whether a 'hit' (search document) should be pruned from an index,
    and if so it returns the hit as a Django object(id=hit_id).

    Args:
        hit: dict object the represents a document as returned from the scan_index
            function. (Contains object id and index.)
        model: the Django model (not object) from which the document was derived.
            Used to get the correct model manager and bulk action.

    Returns:
        an object of type model, with id=hit_id. NB this is not the object
        itself, which by definition may not exist in the underlying database,
        but a temporary object with the document id - which is enough to create
        a 'delete' action.

    """
    hit_id = hit['_id']
    hit_index = hit['_index']
    if model.objects.in_search_queryset(hit_id, index=hit_index):
        logger.debug(
            "%s with id=%s exists in the '%s' index queryset.",
            model, hit_id, hit_index
        )
        return None
    else:
        logger.debug(
            "%s with id=%s does not exist in the '%s' index queryset and will be pruned.",
            model, hit_id, hit_index
        )
        # we don't need the full obj for a delete action, just the id.
        # (the object itself may not even exist.)
        return model(id=hit_id)


def scan_index(index, model):
    """
    Yield all documents of model type in an index.

    This function calls the elasticsearch.helpers.scan function,
    and yields all the documents in the index that match the doc_type
    produced by a specific Django model.

    Args:
        index: string, the name of the index to scan, must be a configured
            index as returned from settings.get_index_names.
        model: a Django model type, used to filter the the documents that
            are scanned.

    Yields each document of type model in index, one at a time.

    """
    # see https://www.elastic.co/guide/en/elasticsearch/reference/current/query-dsl-type-query.html
    query = {"query": {"type": {"value": model._meta.model_name}}}
    client = get_client()
    for hit in helpers.scan(client, index=index, query=query):
        yield hit


def bulk_actions(objects, index, action):
    """
    Yield bulk api 'actions' from a collection of objects.

    The output from this method can be fed in to the bulk
    api helpers - each document returned by get_documents
    is decorated with the appropriate bulk api op_type.

    Args:
        objects: iterable (queryset, list, ...) of SearchDocumentMixin
            objects. If the objects passed in is a generator, then this
                function will yield the results rather than returning them.
        index: string, the name of the index to target - the index name
            is embedded into the return value and is used by the bulk api.
        action: string ['index' | 'update' | 'delete'] - this decides
            how the final document is formatted.

    """
    assert index != '_all', "index arg must be a valid index name. '_all' is a reserved term."
    logger.info("Creating bulk '%s' actions for '%s'", action, index)
    for obj in objects:
        try:
            logger.debug("Appending '%s' action for '%r'", action, obj)
            yield obj.as_search_action(index, action)
        except Exception:
            logger.exception("Unable to create search action for %s", obj)



# -*- coding: utf-8 -*-
"""Access to SEARCH_SETTINGS Django conf.

The SEARCH_SETTINGS dict in the Django conf contains three
major blocks - 'connections', 'indexes' and 'settings'.

This module contains helper functions to extract information
from the settings, as well as validation of settings.

"""
import json
import os

from django.apps import apps
from django.conf import settings

from elasticsearch import Elasticsearch


def get_client(connection='default'):
    """Return configured elasticsearch client."""
    return Elasticsearch(get_connection_string(connection))


def get_settings():
    """Return settings from Django conf."""
    return settings.SEARCH_SETTINGS['settings']


def get_setting(key):
    """Return specific search setting from Django conf."""
    return get_settings()[key]


def get_connection_string(connection='default'):
    """Return index settings from Django conf."""
    return settings.SEARCH_SETTINGS['connections'][connection]


def get_index_config(index):
    """Return index settings from Django conf."""
    return settings.SEARCH_SETTINGS['indexes'][index]


def get_index_names():
    """Return list of the names of all configured indexes."""
    return settings.SEARCH_SETTINGS['indexes'].keys()


def get_index_mapping(index):
    """Return the JSON mapping file for an index.

    Mappings are stored as JSON files in the mappings subdirectory of this
    app. They must be saved as {{index}}.json.

    Args:
        index: string, the name of the index to look for.

    """
    # app_path = apps.get_app_config('elasticsearch_django').path
    mappings_dir = get_setting('mappings_dir')
    filename = '%s.json' % index
    path = os.path.join(mappings_dir, filename)
    with open(path, 'r') as f:
        return json.load(f)


def get_index_models(index):
    """Return list of models configured for a named index.

    Args:
        index: string, the name of the index to look up.

    """
    models = []
    for app_model in get_index_config(index).get('models'):
        app, model = app_model.split('.')
        models.append(apps.get_model(app, model))
    return models


def get_model_indexes(model):
    """Return list of all indexes in which a model is configured.

    A model may be configured to appear in multiple indexes. This function
    will return the names of the indexes as a list of strings. This is
    useful if you want to know which indexes need updating when a model
    is saved.

    Args:
        model: a Django model class.

    """
    indexes = []
    for index in get_index_names():
        for app_model in get_index_models(index):
            if app_model == model:
                indexes.append(index)
    return indexes


def get_document_models():
    """Return dict of index.doc_type: model."""
    mappings = {}
    for i in get_index_names():
        for m in get_index_models(i):
            key = '%s.%s' % (i, m._meta.model_name)
            mappings[key] = m
    return mappings


def get_document_model(index, doc_type):
    """Return dict of index.doc_type: model."""
    return get_document_models().get('%s.%s' % (index, doc_type))



# ######################################################################
# The code below is copied from the Django master - it allows
# for custom encoders, which is what we need to handle the
# serialization of dates and decimals correctly.
#
# Commit: https://github.com/django/django/commit/13c3e5d5a05e9c358d212d154addd703cac3bc66
#
# TODO: deprecate once we upgrade to 1.11(?)
# ######################################################################
from __future__ import absolute_import

import json as base_json

import django.core.serializers.json
import django.contrib.postgres.fields

from psycopg2.extras import Json


class JsonAdapter(Json):

    """Customized psycopg2.extras.Json to allow for a custom encoder."""

    def __init__(self, adapted, dumps=None, encoder=None):
        self.encoder = encoder
        super(JsonAdapter, self).__init__(adapted, dumps=dumps)

    def dumps(self, obj):
        options = {'cls': self.encoder} if self.encoder else {}
        return base_json.dumps(obj, **options)


class JSONField(django.contrib.postgres.fields.JSONField):

    """Subclass of JSONField updated to use DjangoJSONEncoder."""

    # See https://github.com/django/django/blob/master/django/contrib/postgres/fields/jsonb.py

    def __init__(self, verbose_name=None, name=None, **kwargs):
        self.encoder = django.core.serializers.json.DjangoJSONEncoder
        super(JSONField, self).__init__(verbose_name, name, **kwargs)

    def get_prep_value(self, value):
        if value is not None:
            return JsonAdapter(value, encoder=self.encoder)
        return value



# -*- coding: utf-8 -*-
# Generated by Django 1.10 on 2017-01-03 16:11
from __future__ import unicode_literals

from django.db import migrations
import elasticsearch_django.db.fields


class Migration(migrations.Migration):

    dependencies = [
        ('elasticsearch_django', '0004_auto_20161129_1135'),
    ]

    operations = [
        migrations.AlterField(
            model_name='searchquery',
            name='hits',
            field=elasticsearch_django.db.fields.JSONField(help_text=b'The list of meta info for each of the query matches returned.'),
        ),
        migrations.AlterField(
            model_name='searchquery',
            name='query',
            field=elasticsearch_django.db.fields.JSONField(help_text=b'The raw ElasticSearch DSL query.'),
        ),
    ]



# -*- coding: utf-8 -*-
"""Create a search index."""
from elasticsearch_django.management.commands import BaseSearchCommand
from elasticsearch_django.index import create_index


class Command(BaseSearchCommand):

    """Run the management command."""

    help = "Create a new search index using the relevant mapping file."
    description = "Create search index"

    def do_index_command(self, index, **options):
        """Create new search index."""
        return create_index(index)



"""search app models."""
import logging
import time

from django.contrib.auth.models import User
from django.core.cache import cache
from django.db import models
from django.db.models.expressions import RawSQL
from django.utils.timezone import now as tz_now

from .db.fields import JSONField
from .settings import (
    get_client,
    get_model_indexes,
)

logger = logging.getLogger(__name__)


class SearchDocumentManagerMixin(object):

    """
    Model manager mixin that adds search document methods.

    There is one method in this class that must implemented -
    `get_search_queryset`. This must return a queryset that is the
    set of objects to be indexed. This queryset is then converted
    into a generator that emits the objects as JSON documents.

    """

    def get_search_queryset(self, index='_all'):
        """
        Return the dataset used to populate the search index.

        Kwargs:
            index: string, the name of the index we are interested in -
                this allows us to have different sets of objects in
                different indexes. Defaults to '_all', in which case
                all indexes index the same set of objects.

        This must return a queryset object.

        """
        raise NotImplementedError(
            "{} does not implement 'get_search_queryset'.".format(
                self.__class__.__name__
            )
        )

    def in_search_queryset(self, instance_id, index='_all'):
        """
        Return True if an object is part of the search index queryset.

        Sometimes it's useful to know if an object _should_ be indexed. If
        an object is saved, how do you know if you should push that change
        to the search index? The simplest (albeit not most efficient) way
        is to check if it appears in the underlying search queryset.

        NB this method doesn't evaluate the entire dataset, it chains an
        additional queryset filter expression on the end. That's why it's
        important that the `get_search_queryset` method returns a queryset.

        Args:
            instance_id: the id of model object that we are looking for.

        Kwargs:
            index: string, the name of the index in which to check.
                Defaults to '_all'.

        """
        return self.get_search_queryset(index=index).filter(id=instance_id).exists()

    def from_search_query(self, search_query):
        """
        Return queryset of objects from SearchQuery.results, **in order**.

        EXPERIMENTAL: this will only work with results from a single index,
        with a single doc_type - as we are returning a single QuerySet.

        This method takes the hits JSON and converts that into a queryset
        of all the relevant objects. The key part of this is the ordering -
        the order in which search results are returned is based on relevance,
        something that only ES can calculate, and that cannot be replicated
        in the database.

        It does this by adding custom SQL which annotates each record with
        the score from the search 'hit'. This is brittle, caveat emptor.

        The RawSQL clause is in the form:

            SELECT CASE {{model}}.id WHEN {{id}} THEN {{score}} END

        The "WHEN x THEN y" is repeated for every hit. The resulting SQL, in
        full is like this:

            SELECT "freelancer_freelancerprofile"."id",
                (SELECT CASE freelancer_freelancerprofile.id
                    WHEN 25 THEN 1.0
                    WHEN 26 THEN 1.0
                    [...]
                    ELSE 0
                END) AS "search_score"
            FROM "freelancer_freelancerprofile"
            WHERE "freelancer_freelancerprofile"."id" IN (25, 26, [...])
            ORDER BY "search_score" DESC

        It should be very fast, as there is no table lookup, but there is an
        assumption at the heart of this, which is that the search query doesn't
        contain the entire database - i.e. that it has been paged. (ES itself
        caps the results at 10,000.)

        """
        hits = search_query.hits
        score_sql = self._raw_sql([(h['id'], h['score']) for h in hits])
        rank_sql = self._raw_sql([(hits[i]['id'], i) for i in xrange(len(hits))])
        return (
            self.get_queryset()
            .filter(id__in=[h['id'] for h in hits])
            # add the query relevance score
            .annotate(search_score=RawSQL(score_sql, ()))
            # add the ordering number (0-based)
            .annotate(search_rank=RawSQL(rank_sql, ()))
            .order_by('search_rank')
        )

    def _raw_sql(self, values):
        """Prepare SQL statement consisting of a sequence of WHEN .. THEN statements."""
        when_ = lambda (x, y): "WHEN {} THEN {}".format(x, y)
        when_clauses = ' '.join([when_(h) for h in values])
        table_name = self.model._meta.db_table
        return "SELECT CASE {}.id {} ELSE 0 END".format(table_name, when_clauses)


class SearchDocumentMixin(object):

    """
    Mixin used by models that are indexed for ES.

    This mixin defines the interface exposed by models that
    are indexed ready for ES. The only method that needs
    implementing is `as_search_document`.

    """

    @property
    def search_indexes(self):
        """Return the list of indexes for which this model is configured."""
        return get_model_indexes(self.__class__)

    @property
    def search_document_cache_key(self):
        """Key used for storing search docs in local cache."""
        return 'elasticsearch_django:{}.{}.{}'.format(
            self._meta.app_label,
            self._meta.model_name,
            self.id
        )

    @property
    def search_doc_type(self):
        """Return the doc_type used for the model."""
        return self._meta.model_name

    def as_search_document(self, index='_all'):
        """
        Return the object as represented in a named index.

        This is named to avoid confusion - if it was `get_search_document`,
        which would be the logical name, it would not be clear whether it
        referred to getting the local representation of the search document,
        or actually fetching it from the index.

        Kwargs:
            index: string, the name of the index in which the object is to
                appear - this allows different representations in different
                indexes. Defaults to '_all', in which case all indexes use
                the same search document structure.

        Returns a dictionary.

        """
        raise NotImplementedError(
            "{} does not implement 'get_search_document'.".format(self.__class__.__name__)
        )

    def as_search_action(self, index, action):
        """
        Return an object as represented in a bulk api operation.

        Bulk API operations have a very specific format. This function will
        call the standard `as_search_document` method on the object and then
        wrap that up in the correct format for the action specified.

        https://www.elastic.co/guide/en/elasticsearch/reference/current/docs-bulk.html

        Args:
            index: string, the name of the index in which the action is to
                be taken. Bulk operations are only every carried out on a single
                index at a time.
            action: string ['index' | 'update' | 'delete'] - this decides
                how the final document is formatted.

        Returns a dictionary.

        """
        assert action in ('index', 'update', 'delete'), ("Action must be 'index', 'update' or 'delete'.")  # noqa

        document = {
            '_index': index,
            '_type': self.search_doc_type,
            '_op_type': action,
            '_id': self.id,
        }

        if action == 'index':
            document['_source'] = self.as_search_document(index)
        elif action == 'update':
            document['doc'] = self.as_search_document(index)
        return document

    def fetch_search_document(self, index):
        """Fetch the object's document from a search index by id."""
        assert self.id, ("Object must have a primary key before being indexed.")
        client = get_client()
        return client.get(
            index=index,
            doc_type=self.search_doc_type,
            id=self.id
        )

    def update_search_index(self, action, index='_all', force=False):
        """
        Update the object in a remote index.

        This method is used to run index, update and delete actions on
        the index. We have a single method rather than individual methods
        in order to preserve the semantics of what we are doing. Irrespective
        of whether we are creating a new document, or deleting an existing
        document, from the point of view of the client we are updating the
        state of the remote search index.

        Before attempting to update the index, this method will check to see
        if the object is in the search queryset to determine whether it should
        be indexed at all.

        Args:
            action: string ['index' | 'update' | 'delete'] - the action to
                take - resolves to a POST, PUT, DELETE action

        Kwargs:
            index: string, the name of the index to update. Defaults to '_all',
                which is a reserved ES term meaning all indexes. In this case we
                use the config to look up all configured indexes for the model.
            force: bool, if True then ignore caching and force the update. Defaults
                to False.

        NB In reality we only support 'index' and 'delete' - 'update' is really
        a PATCH operation, updating partial documents in the search index - and
        we don't currently support this - we only ever update the entire document.

        Returns the HTTP response.

        """
        assert action in ('index', 'update', 'delete'), ("Action must be 'index', 'update' or 'delete'.")  # noqa
        assert self.id, ("Object must have a primary key before being indexed.")

        if not self.objects.in_search_queryset(self.id, index=index):
            logger.debug(
                "%r is not in the source queryset for '%s', aborting update.",
                self, index
            )
            return None

        if action == 'update':
            logger.warn(
                "'update' action is unsupported - switching to 'index' instead."
            )
            action = 'index'

        # use all configured indexes if none was passed in, else whatever we were given
        indexes = self.search_indexes if index == '_all' else [index]
        responses = []
        for i in indexes:
            responses.append(self._do_search_action(i, action, force=force))
        return responses

    def _do_search_action(self, index, action, force=False):
        """
        Call the relevant api function.

        This is where the API itself is used (for single document actions),
        but it shouldn't be used directly - the public method is `update_search_index`.

        Args:
            index: string, the name of the index to update.
            action: string, must be either 'index' or 'delete'.
            force: bool, if True then ignore cache and force the update

        Returns the HTTP response from the API call.

        NB this contains one of the core assumptions - that the model name is
        used as the search index document type name.

        """
        assert self.id, ("Object must have a primary key before being indexed.")
        assert action in ('index', 'delete'), (
            "Search action '{}' is invalid; must be 'index' or 'delete'.".format(action)
        )
        client = get_client()
        cache_key = self.search_document_cache_key
        if action == 'index':
            # if the locally cached search doc is the same as the new one,
            # then don't bother pushing to ES.
            new_doc = self.as_search_document(index)
            if not force:
                cached_doc = cache.get(cache_key)
                if new_doc == cached_doc:
                    logger.debug("Search document for %r is unchanged, ignoring update.", self)
                    return []
            cache.set(cache_key, new_doc, timeout=60)  # TODO: remove hard-coded timeout
            return client.index(
                index=index,
                doc_type=self.search_doc_type,
                body=new_doc,
                id=self.id
            )

        if action == 'delete':
            cache.delete(cache_key)
            return client.delete(
                index=index,
                doc_type=self.search_doc_type,
                id=self.id
            )


class SearchQuery(models.Model):

    """
    Model used to capture ES queries and responses.

    For low-traffic sites it's useful to be able to replay
    searches, and to track how a user filtered and searched.
    This model can be used to store a search query and meta
    information about the results (document type, id and score).

    >>> from elasticsearch_dsl import Search
    >>> search = Search(using=client)
    >>> sq = SearchQuery.execute(search).save()

    """

    user = models.ForeignKey(
        User,
        related_name='search_queries',
        blank=True, null=True,
        help_text="The user who made the search query (nullable)."
    )
    index = models.CharField(
        max_length=100,
        default='_all',
        help_text="The name of the ElasticSearch index(es) being queried."
    )
    query = JSONField(
        help_text="The raw ElasticSearch DSL query."
    )
    hits = JSONField(
        help_text="The list of meta info for each of the query matches returned."
    )
    total_hits = models.IntegerField(
        default=0,
        help_text="Total number of matches found for the query (!= the hits returned)."
    )
    reference = models.CharField(
        max_length=100,
        default='',
        blank=True,
        help_text="Custom reference used to identify and group related searches."
    )
    executed_at = models.DateTimeField(
        help_text="When the search was executed - set via execute() method."
    )
    duration = models.FloatField(
        help_text="Time taken to execute the search itself, in seconds."
    )

    class Meta:
        app_label = 'elasticsearch_django'
        verbose_name = "Search query"
        verbose_name_plural = "Search queries"

    def __unicode__(self):
        return (
            u"Query (id=%s) run against index '%s'" % (
                self.id, self.index
            )
        )

    def __repr__(self):
        return (
            u"<QueryLog id=%s user=%s index='%s' total_hits=%i >" % (
                self.id, self.user, self.index, self.total_hits
            )
        )

    @classmethod
    def execute(cls, search, user=None, reference=None, save=True):
        """
        Create a new SearchQuery instance and execute a search against ES.

        Args:
            search: elasticsearch.search.Search object, that internally contains
                the connection and query; this is the query that is executed. All
                we are doing is logging the input and parsing the output.
            user: Django User object, the person making the query - used for logging
                purposes. Can be null.
            reference: string, can be anything you like, used for identification,
                grouping purposes.
            save: bool, if True then save the new object immediately, can be
                overridden to False to prevent logging absolutely everything.
                Defaults to True

        """
        start = time.time()
        response = search.execute()
        duration = time.time() - start
        log = SearchQuery(
            user=user,
            index=', '.join(search._index or ['_all'])[:100],  # field length restriction
            query=search.to_dict(),
            hits=[h.meta.to_dict() for h in response.hits],
            total_hits=response.hits.total,
            reference=reference or '',
            executed_at=tz_now(),
            duration=duration
        )
        return log.save() if save else log

    def save(self, **kwargs):
        """Save and return the object (for chainging)."""
        super(SearchQuery, self).save(**kwargs)
        return self

    def _extract_set(self, _property):
        return [] if self.hits is None else (
            list(set([h[_property] for h in self.hits]))
        )

    @property
    def doc_types(self):
        """List of doc_types extracted from hits."""
        return self._extract_set('doc_type')

    @property
    def max_score(self):
        """The max relevance score in the returned page."""
        return max(self._extract_set('score') or [0])

    @property
    def min_score(self):
        """The min relevance score in the returned page."""
        return min(self._extract_set('score') or [0])

    @property
    def object_ids(self):
        """List of model ids extracted from hits."""
        return self._extract_set('id')

    @property
    def page_slice(self):
        """Return the query from:size tuple (0-based)."""
        return None if self.query is None else (
            self.query.get('from', 0),
            self.query.get('size', 10)
        )

    @property
    def page_from(self):
        """1-based index of the first hit in the returned page."""
        return 0 if self.page_size == 0 else self.page_slice[0] + 1

    @property
    def page_to(self):
        """1-based index of the last hit in the returned page."""
        return 0 if self.page_size == 0 else self.page_from + self.page_size - 1

    @property
    def page_size(self):
        """The number of hits returned in this specific page."""
        return 0 if self.hits is None else len(self.hits)



# -*- coding: utf-8 -*-
"""Search3 app model tests."""
import datetime
import decimal
import mock

from django.db.models import Model
from django.test import TestCase
from django.utils.timezone import now as tz_now

from elasticsearch_dsl.search import Search

from ..models import (
    SearchDocumentMixin,
    SearchDocumentManagerMixin,
    SearchQuery,
)
from ..tests import (
    TestModel,
    TestModelManager,
    SEARCH_DOC,
)


class SearchDocumentMixinTests(TestCase):

    """Tests for the SearchDocumentMixin."""

    @mock.patch('elasticsearch_django.models.get_model_indexes')
    def test_search_indexes(self, mock_indexes):
        """Test the search_indexes function."""
        mock_indexes.return_value = 'foo'
        obj = TestModel()
        self.assertEqual(obj.search_indexes, 'foo')
        mock_indexes.assert_called_once_with(TestModel)

    def test_as_search_document(self):
        """Test the as_search_document method."""
        obj = SearchDocumentMixin()
        self.assertRaises(NotImplementedError, obj.as_search_document)

    def test_as_search_action(self):
        """Test the as_search_action method."""
        obj = TestModel()
        self.assertEqual(
            obj.as_search_action(index='foo', action='index'),
            {
                '_index': 'foo',
                '_type': 'testmodel',
                '_op_type': 'index',
                '_id': None,
                '_source': SEARCH_DOC
            }
        )

        self.assertEqual(
            obj.as_search_action(index='foo', action='update'),
            {
                '_index': 'foo',
                '_type': 'testmodel',
                '_op_type': 'update',
                '_id': None,
                'doc': SEARCH_DOC
            }
        )

        self.assertEqual(
            obj.as_search_action(index='foo', action='delete'),
            {
                '_index': 'foo',
                '_type': 'testmodel',
                '_op_type': 'delete',
                '_id': None
            }
        )

    @mock.patch('elasticsearch_django.models.get_client')
    def test_fetch_search_document(self, mock_client):
        """Test the fetch_search_document method."""
        obj = TestModel()
        # obj has no id
        self.assertRaises(AssertionError, obj.fetch_search_document, index='foo')

        # should now call the ES get method
        obj.id = 1
        response = obj.fetch_search_document(index='foo')
        mock_get = mock_client.return_value.get
        mock_get.assert_called_once_with(
            index='foo',
            doc_type=obj._meta.model_name,
            id=obj.id
        )
        self.assertEqual(response, mock_get.return_value)

    @mock.patch('elasticsearch_django.tests.TestModel.objects')
    @mock.patch.object(TestModel, 'search_indexes', new_callable=mock.PropertyMock)
    @mock.patch.object(TestModel, '_do_search_action')
    def test_update_search_index(self, mock_do_search, mock_indexes, mock_manager):
        """Test the update_search_index method."""
        obj = TestModel()

        # invalid action 'foo'
        self.assertRaises(AssertionError, obj.update_search_index, action='foo')
        mock_manager.assert_not_called()

        # valid action, but no id
        self.assertRaises(AssertionError, obj.update_search_index, action='index')
        mock_manager.assert_not_called()

        obj.id = 1
        # the object is not in the search queryset, should **not** call the update
        mock_manager.in_search_queryset.return_value = False
        response = obj.update_search_index(action='index')
        mock_manager.in_search_queryset.assert_called_once_with(obj.id, index='_all')
        self.assertIsNone(response)

        # check that 'update' actions are converted to 'index'
        mock_manager.reset_mock()
        mock_manager.in_search_queryset.return_value = True
        response = obj.update_search_index(action='update', index='foo')
        mock_manager.in_search_queryset.assert_called_once_with(obj.id, index='foo')
        mock_do_search.assert_called_once_with('foo', 'index', force=False)

        # check that 'index' actions go through as 'index'
        mock_manager.reset_mock()
        mock_do_search.reset_mock()
        response = obj.update_search_index(action='index', index='bar')
        mock_manager.in_search_queryset.assert_called_once_with(obj.id, index='bar')
        mock_do_search.assert_called_once_with('bar', 'index', force=False)

        mock_manager.reset_mock()
        mock_do_search.reset_mock()
        mock_indexes.return_value = ['foo', 'bar']
        response = obj.update_search_index(action='index')
        mock_manager.in_search_queryset.assert_called_once_with(obj.id, index='_all')
        mock_do_search.assert_has_calls([
            mock.call('foo', 'index', force=False),
            mock.call('bar', 'index', force=False)
        ])

    @mock.patch('elasticsearch_django.models.cache')
    @mock.patch('elasticsearch_django.models.get_client')
    def test__do_search_action(self, mock_client, mock_cache):
        """Test the _do_search_action method."""
        obj = TestModel()

        # valid action, but not id
        self.assertRaises(AssertionError, obj._do_search_action, index='foo', action='index')
        # id is good, but invalid action
        obj.id = 1
        self.assertRaises(AssertionError, obj._do_search_action, index='foo', action='foobar')

        obj._do_search_action(index='foo', action='index')
        mock_index = mock_client.return_value.index
        mock_index.assert_called_once_with(
            index='foo',
            doc_type=obj._meta.model_name,
            body=SEARCH_DOC,
            id=obj.id
        )

        obj._do_search_action(index='foo', action='delete')
        mock_delete = mock_client.return_value.delete
        mock_delete.assert_called_once_with(
            index='foo',
            doc_type=obj._meta.model_name,
            id=obj.id
        )

        # test the caching - cahce hit, index should not be called
        mock_client.reset_mock()
        mock_index.reset_mock()
        mock_cache.get.return_value = obj.as_search_document('foo')
        obj._do_search_action(index='foo', action='index')
        mock_index.assert_not_called()

        # cache hit, but with force=True - should call the index
        mock_client.reset_mock()
        mock_index.reset_mock()
        obj._do_search_action(index='foo', action='index', force=True)
        mock_index.assert_called_once_with(
            index='foo',
            doc_type=obj._meta.model_name,
            body=SEARCH_DOC,
            id=obj.id
        )


class SearchDocumentManagerMixinTests(TestCase):

    """Tests for the SearchDocumentManagerMixin."""

    def test_get_search_queryset(self):
        """Test the get_search_queryset method."""
        obj = SearchDocumentManagerMixin()
        self.assertRaises(NotImplementedError, obj.get_search_queryset)

    @mock.patch.object(TestModelManager, 'get_search_queryset')
    def test_in_search_queryset(self, mock_qs):
        """Test the in_search_queryset method."""
        obj = TestModel(id=1)
        TestModel.objects.in_search_queryset(obj.id)
        mock_qs.assert_called_once_with(index='_all')
        mock_qs.return_value.filter.assert_called_once_with(id=1)
        mock_qs.return_value.filter.return_value.exists.assert_called_once_with()

    def test__raw_sql(self):
        """Test the _raw_sql method."""
        self.assertEqual(
            TestModel.objects._raw_sql(((1, 2), (3, 4))),
            "SELECT CASE elasticsearch_django_testmodel.id WHEN 1 THEN 2 WHEN 3 THEN 4 ELSE 0 END"
        )

    @mock.patch('django.db.models.query.QuerySet')
    def test_from_search_query(self, mock_qs):
        """Test the from_search_query method."""
        sq = SearchQuery(hits=[{'id': 1, 'score': 1}, {'id': 2, 'score': 2}])
        qs = TestModel.objects.from_search_query(sq)
        self.assertEqual(
            str(qs.query),
            'SELECT "elasticsearch_django_testmodel"."id", '
            '(SELECT CASE elasticsearch_django_testmodel.id WHEN 1 THEN 1 WHEN 2 THEN 2 ELSE 0 END) '  # noqa
            'AS "search_score", (SELECT CASE elasticsearch_django_testmodel.id WHEN 1 THEN 0 WHEN 2 '  # noqa
            'THEN 1 ELSE 0 END) AS "search_rank" FROM "elasticsearch_django_testmodel" WHERE '
            '"elasticsearch_django_testmodel"."id" IN (1, 2) ORDER BY "search_rank" ASC'
        )


class SearchQueryTests(TestCase):

    """Tests for the SearchQuery model."""

    hits = [
        {'id': 1, 'doc_type': 'foo'},
        {'id': 2, 'doc_type': 'foo'},
        {'id': 3, 'doc_type': 'bar'},
    ]

    def test__extract_set(self):
        """Test the _extract_set method."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj._extract_set('id')), set([1, 2, 3]))

    def test_doc_types(self):
        """Test the doc_types property."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj.doc_types), set(['foo', 'bar']))

    def test_object_ids(self):
        """Test the object_ids property."""
        obj = SearchQuery(hits=SearchQueryTests.hits)
        self.assertEqual(set(obj.object_ids), set([1, 2, 3]))

    def test_save(self):
        """Try saving unserializable JSON."""
        today = datetime.date.today()
        sq = SearchQuery(
            user=None,
            index='foo',
            query={'today': today},
            hits={'hits': decimal.Decimal('1.0')},
            total_hits=100,
            reference='bar',
            executed_at=tz_now(),
            duration=0
        )
        sq.save()
        sq.refresh_from_db()
        # invalid JSON values will have been converted
        self.assertEqual(sq.query['today'], today.isoformat())
        self.assertEqual(sq.hits['hits'], '1.0')

    def test_paging(self):
        """Test the paging properties."""
        sq = SearchQuery()
        self.assertEqual(sq.page_slice, None)

        # no hits, so should all be 0
        sq.query = {'from': 0, 'size': 25}
        self.assertEqual(sq.page_slice, (0, 25))
        self.assertEqual(sq.page_from, 0)
        self.assertEqual(sq.page_to, 0)
        self.assertEqual(sq.page_size, 0)

        # three hits
        sq.hits = [1, 2, 3]  # random list of size = 3
        sq.query = {'from': 0, 'size': 25}
        self.assertEqual(sq.page_from, 1)
        self.assertEqual(sq.page_to, 3)
        self.assertEqual(sq.page_size, 3)

    def test_scores(self):
        """Test the max/min properties."""
        sq = SearchQuery()
        self.assertEqual(sq.max_score, 0)
        self.assertEqual(sq.min_score, 0)

        sq.hits = [{'score': 1}, {'score': 2}]
        self.assertEqual(sq.max_score, 2)
        self.assertEqual(sq.min_score, 1)

    @mock.patch('elasticsearch_django.models.tz_now')
    @mock.patch.object(Search, 'execute')
    @mock.patch.object(Model, 'save')
    def test_execute(self, mock_save, mock_execute, mock_now):
        """Test the execute class method."""
        search = Search()
        sq = SearchQuery.execute(search)
        self.assertEqual(sq.user, None)
        self.assertEqual(sq.index, '_all')
        self.assertEqual(sq.query, search.to_dict())
        self.assertEqual(sq.hits, [])
        self.assertEqual(sq.total_hits, mock_execute.return_value.hits.total)
        self.assertEqual(sq.reference, '')
        self.assertTrue(sq.duration > 0)
        self.assertEqual(sq.executed_at, mock_now.return_value)
        mock_save.assert_called_once_with()

        # try without saving
        mock_save.reset_mock()
        sq = SearchQuery.execute(search, save=False)
        mock_save.assert_not_called()



# -*- coding: utf-8 -*-
from __future__ import unicode_literals

from django.db import models, migrations


class Migration(migrations.Migration):

    dependencies = [
        ('elasticsearch_django', '0001_initial'),
    ]

    operations = [
        migrations.AddField(
            model_name='searchquery',
            name='duration',
            field=models.FloatField(default=0, help_text=b'Time taken to execute the search itself, in seconds.'),
            preserve_default=False,
        ),
    ]



# -*- coding: utf-8 -*-
# search.tests package identifier
from django.db import models

from ..models import (
    SearchDocumentMixin,
    SearchDocumentManagerMixin
)


SEARCH_DOC = {"foo": "bar"}


class TestModelManager(SearchDocumentManagerMixin, models.Manager):
    pass


class TestModel(SearchDocumentMixin, models.Model):

    """Model class for use in tests."""

    class Meta:
        # should prevent db errors during tests
        managed = False

    objects = TestModelManager()

    def as_search_document(self, index='_all'):
        return SEARCH_DOC



# -*- coding: utf-8 -*-
from contextlib import contextmanager

from .settings import get_settings


@contextmanager
def disable_search_updates():
    """
    Context manager used to temporarily disable auto_sync.

    This is useful when performing bulk updates on objects - when
    you may not want to flood the indexing process.

    >>> with disable_search_updates():
    ...     for obj in model.objects.all():
    ...     obj.save()

    """
    _settings = get_settings()
    _sync = _settings['auto_sync']
    _settings['auto_sync'] = False
    yield
    _settings['auto_sync'] = _sync



# -*- coding: utf-8 -*-
"""Search3 app management commands tests."""
import mock

from django.test import TestCase

from elasticsearch.exceptions import TransportError

from ..management.commands import (
    BaseSearchCommand,
    create_search_index,
    delete_search_index,
    prune_search_index,
    update_search_index,
    rebuild_search_index
)


class BaseSearchCommandTests(TestCase):

    """Tests for the elasticsearch_django management commands base command."""

    @mock.patch('elasticsearch_django.management.commands.logger')
    @mock.patch.object(BaseSearchCommand, 'do_index_command')
    def test_handle(self, mock_do, mock_log):
        """Test the main handle method calls do_index_command."""
        obj = BaseSearchCommand()
        obj.handle(indexes=['foo', 'bar'])
        # this should have called the do_index_command twice
        mock_do.assert_has_calls([mock.call('foo'), mock.call('bar')])
        mock_do.reset_mock()
        mock_do.side_effect = TransportError(123, "oops", {'error': {'reason': 'no idea'}})
        obj.handle(indexes=['baz'])
        mock_do.assert_called_once_with('baz')
        mock_log.warn.assert_called_once()


class NamedCommandTests(TestCase):

    """Test each named command."""

    @mock.patch('elasticsearch_django.management.commands.create_search_index.create_index')
    def test_create_search_index(self, mock_create):
        """Test the create_search_index command."""
        cmd = create_search_index.Command()
        cmd.do_index_command('foo')
        mock_create.assert_called_once_with('foo')

    @mock.patch('elasticsearch_django.management.commands.delete_search_index.delete_index')
    def test_delete_search_index(self, mock_delete):
        """Test the delete_search_index command."""
        cmd = delete_search_index.Command()
        retval = cmd.do_index_command('foo', interactive=False)  # True would hang the tests
        self.assertEqual(retval, mock_delete.return_value)
        mock_delete.assert_called_once_with('foo')
        mock_delete.reset_mock()

        # mock out thw raw_input so the test doesn't hang
        with mock.patch.object(delete_search_index.Command, '_confirm_action') as mock_confirm:
            mock_confirm.return_value = False
            retval = cmd.do_index_command('foo', interactive=True)
            mock_delete.assert_not_called()
            self.assertIsNone(retval)

    @mock.patch('elasticsearch_django.management.commands.prune_search_index.prune_index')
    def test_prune_search_index(self, mock_prune):
        """Test the prune_search_index command."""
        cmd = prune_search_index.Command()
        cmd.do_index_command('foo')
        mock_prune.assert_called_once_with('foo')

    @mock.patch('elasticsearch_django.management.commands.update_search_index.update_index')
    def test_update_search_index(self, mock_update):
        """Test the update_search_index command."""
        cmd = update_search_index.Command()
        cmd.do_index_command('foo')
        mock_update.assert_called_once_with('foo')

    @mock.patch('elasticsearch_django.management.commands.rebuild_search_index.delete_index')
    @mock.patch('elasticsearch_django.management.commands.rebuild_search_index.create_index')
    @mock.patch('elasticsearch_django.management.commands.rebuild_search_index.update_index')
    def test_update_search_index(self, mock_update, mock_create, mock_delete):
        """Test the rebuild_search_index command."""
        cmd = rebuild_search_index.Command()
        cmd.do_index_command('foo', interactive=False)  # True would hang the tests
        mock_delete.assert_called_once_with('foo')
        mock_create.assert_called_once_with('foo')
        mock_update.assert_called_once_with('foo')



# -*- coding: utf-8 -*-
"""Update all documents in a search index."""
from elasticsearch_django.management.commands import BaseSearchCommand
from elasticsearch_django.index import update_index


class Command(BaseSearchCommand):

    """Run the management command."""

    help = "Update all documents in a search index."
    description = "Update search index."

    def do_index_command(self, index, **options):
        """Update search index."""
        return update_index(index)



# -*- coding: utf-8 -*-
"""Search3 app index function tests."""
import mock

from django.test import TestCase

from ..index import (
    create_index,
    update_index,
    delete_index,
    prune_index,
    _prune_hit,
    scan_index,
    bulk_actions
)
from ..tests import (
    TestModel,
    TestModelManager
)


class IndexFunctionTests(TestCase):

    """Test index functions."""

    @mock.patch('elasticsearch_django.index.get_client')
    @mock.patch('elasticsearch_django.index.get_index_mapping')
    def test_create_index(self, mock_mapping, mock_client):
        """Test the create_index function."""
        mock_client.return_value = mock.Mock()
        create_index('foo')
        mock_client.assert_called_once_with()
        mock_mapping.assert_called_once_with('foo')
        mock_client.return_value.indices.create.assert_called_once_with(
            index='foo', body=mock_mapping.return_value
        )

    from django.db.models.query import QuerySet

    @mock.patch.object(QuerySet, 'iterator')
    @mock.patch('elasticsearch_django.index.get_client')
    @mock.patch('elasticsearch_django.index.bulk_actions')
    @mock.patch('elasticsearch_django.index.get_index_models')
    @mock.patch('elasticsearch.helpers.bulk')
    def test_update_index(self, mock_bulk, mock_models, mock_actions, mock_client, mock_qs):
        """Test the update_index function."""
        mock_foo = mock.Mock
        mock_foo.search_doc_type = mock.PropertyMock(return_value='bar')
        mock_foo.objects = mock.PropertyMock(return_value=mock.Mock())
        mock_models.return_value = [mock_foo]
        responses = update_index('foo')
        self.assertEqual(responses, [mock_bulk.return_value])

    @mock.patch('elasticsearch_django.index.get_client')
    def test_delete_index(self, mock_client):
        """Test the delete_index function."""
        delete_index('foo')
        mock_client.assert_called_once()
        mock_client.return_value.indices.delete.assert_called_once_with(index='foo')

    @mock.patch('elasticsearch_django.index.helpers')
    @mock.patch('elasticsearch_django.index.scan_index')
    @mock.patch('elasticsearch_django.index._prune_hit')
    @mock.patch('elasticsearch_django.index.bulk_actions')
    @mock.patch('elasticsearch_django.index.get_index_models')
    @mock.patch('elasticsearch_django.index.get_client')
    def test_prune_index(
        self, mock_client, mock_models,
        mock_actions, mock_prune, mock_scan, mock_helpers
    ):
        """Test the prune_index function."""

        # this forces one single evaluation of the outer and inner for loop
        mock_models.return_value = [TestModel]
        mock_scan.return_value = ["hit"]

        # _prune_hit returns an object, so bulk should be called
        mock_prune.return_value = TestModel()
        # should return a list with one item in it
        self.assertEqual(prune_index('foo'), [mock_helpers.bulk.return_value])
        # should have called actions and bulk once each
        mock_actions.assert_called_once()
        mock_helpers.bulk.assert_called_once()

        mock_actions.reset_mock()
        mock_helpers.bulk.reset_mock()
        # if there are no objects to prune
        mock_prune.return_value = None
        # should return an empty list
        self.assertEqual(prune_index('foo'), [])
        # shouldn't call either actions or bulk (as there's no need)
        mock_actions.assert_not_called()
        mock_helpers.bulk.assert_not_called()

    @mock.patch.object(TestModelManager, 'in_search_queryset')
    def test__prune_hit(self, mock_qs):
        """Test the _prune_hit function."""
        hit = {"_id": 1, "_index": "foo"}
        mock_qs.return_value = True
        self.assertIsNone(_prune_hit(hit, TestModel))

        mock_qs.return_value = False
        # should now return an instance of TestModel
        obj = _prune_hit(hit, TestModel)
        self.assertIsInstance(obj, TestModel)
        self.assertEqual(obj.id, hit['_id'])

    @mock.patch('elasticsearch_django.index.get_client')
    @mock.patch('elasticsearch_django.index.helpers')
    def test_scan_index(self, mock_helpers, mock_client):
        """Test the scan_index function."""
        query = {'query': {'type': {'value': 'testmodel'}}}
        # mock_helpers.scan.return_value = ['foo', 'bar']
        # cast to list to force evaluation of the generator
        response = list(scan_index('foo', TestModel))
        mock_helpers.scan.assert_called_once_with(
            mock_client.return_value,
            query=query,
            index='foo'
        )
        self.assertEqual(response, list(mock_helpers.scan.return_value))

    @mock.patch.object(TestModel, 'as_search_action')
    def test_bulk_actions(self, mock_action):
        """Test the bulk_actions function."""
        # cannot pass in in '_all' as the bulk_actions
        with self.assertRaises(AssertionError):
            list(bulk_actions([], '_all', 'index'))

        mock_action.return_value = 'foo'
        objects = [TestModel(), TestModel()]

        self.assertEqual(
            list(bulk_actions(objects, 'foo', 'update')),
            ['foo', 'foo']
        )

        # now let's add in a bad object, and check we still get the good one
        self.assertEqual(
            list(bulk_actions([TestModel(), "bad"], 'foo', 'update')),
            ['foo']
        )
